@@ -1,8 +1,8 @@
 import warnings
-import collections
-from collections import namedtuple
+import collections.abc
+from collections import OrderedDict, namedtuple
 import mplhep._deprecate as deprecate
-from typing import Union
+from typing import Union, List, Optional, Dict, Any, TYPE_CHECKING
 
 import matplotlib as mpl
 import matplotlib.pyplot as plt
@@ -12,7 +12,15 @@ from matplotlib.offsetbox import AnchoredText
 from mpl_toolkits.axes_grid1 import make_axes_locatable, axes_size
 
 from .error_estimation import poisson_interval
-from .utils import get_histogram_axes_title
+from .utils import (
+    get_histogram_axes_title,
+    process_histogram_parts,
+    get_plottable_protocol_bins,
+    hist_object_handler,
+)
+
+if TYPE_CHECKING:
+    from numpy.typing import ArrayLike
 
 # mpl updated to new methods
 # from packaging import version
@@ -24,17 +32,42 @@ FillArtists = namedtuple("FillArtists", "fill_between")
 ErrorBarArtists = namedtuple("ErrorBarArtists", "errorbar")
 ColormeshArtists = namedtuple("ColormeshArtists", "pcolormesh cbar")
 
+
 Hist1DArtists = Union[StepArtists, FillArtists, ErrorBarArtists]
 Hist2DArtists = ColormeshArtists
 
 
+def get_density(
+    h, density: bool = True, binwnorm: Optional[float] = None, *, bins, hists
+):
+    if density and binwnorm is not None:
+        raise ValueError("Can only calculate density or binwnorm")
+    per_hist_norm = np.sum(h, axis=(1 if h.ndim > 1 else 0))
+    if binwnorm is not None:
+        overallnorm = binwnorm * per_hist_norm
+    else:
+        overallnorm = np.ones(len(hists))
+    binnorms = np.outer(overallnorm, np.ones_like(bins[:-1]))
+    binnorms /= np.outer(np.diff(bins), per_hist_norm).T
+    if binnorms.ndim == 2 and len(binnorms) == 1:  # Unwrap if [[1,2,3]]
+        binnorms = binnorms[0]
+    return binnorms
+
+
+def get_stack(_h, hists):
+    if len(hists) > 1:
+        return np.cumsum(_h, axis=0)
+    else:
+        return _h
+
+
 ########################################
 # Histogram plotter
 def histplot(
     H,  # Histogram object, tuple or array
     bins=None,  # Bins to be supplied when h is a value array or iterable of arrays
     *,
-    yerr=None,
+    yerr: "ArrayLike | bool | None" = None,
     w2=None,
     w2method=None,
     stack=False,
@@ -55,7 +88,8 @@ def histplot(
             Histogram object with containing values and optionally bins. Can be:
 
             - `np.histogram` tuple
-            - `boost_histogram` histogram object
+            - PlottableProtocol histogram object
+            - `boost_histogram` classic (<0.13) histogram object
             - raw histogram values, provided `bins` is specified.
 
         bins : iterable, optional
@@ -66,9 +100,9 @@ def histplot(
             - shape(N) array of for one sided errors or list thereof
             - shape(Nx2) array of for two sided errors or list thereof
         w2 : iterable, optional
-            Sum of the histogram weights squared for poissonian interval error caclulation
+            Sum of the histogram weights squared for poissonian interval error calculation
         w2method: callable, optional
-            Function calulating CLs with signature ``low, high = fcn(w, w2)``. Here
+            Function calculating CLs with signature ``low, high = fcn(w, w2)``. Here
             ``low`` and ``high`` are given in absolute terms, not relative to w.
             Default is ``None``. If w2 has integer values (likely to be data) poisson
             interval is calculated, otherwise the resulting error is symmetric ``sqrt(w2)``.
@@ -116,112 +150,43 @@ def histplot(
     _err_message = f"Select 'histtype' from: {_allowed_histtype}"
     assert histtype in _allowed_histtype, _err_message
 
-    def hist_object_handler(h, check_only=False):
-        if hasattr(h, "axes") and hasattr(h, "to_numpy"):
-            # Boost histogram / Hist compat
-            # TODO: support different storages
-            if len(h.axes) != 1:
-                raise ValueError("Must have only 1 axis")
-            # FIXME x_axes_label = get_histogram_axes_title(h.axes[0])
-            h, bins = h.to_numpy()
+    hists = list(process_histogram_parts(H, bins))
+    final_bins = get_plottable_protocol_bins(hists[0].axes[0])
 
-        elif hasattr(h, "to_numpy"):
-            # Generic (possibly Uproot 4)
-            _tup = h.to_numpy(flow=False)
-            if len(_tup) != 2:
-                raise ValueError("to_numpy() method not understood")
-            else:
-                h, bins = _tup
-
-        elif hasattr(h, "numpy"):
-            # uproot/TH1
-            _tup = h.numpy()
-            if len(_tup) != 2:
-                raise ValueError("numpy() method not understood")
-            else:
-                h, bins = _tup
-
-        elif isinstance(h, tuple):
-            # Numpy histogram tuple
-            h, bins = h
-
-        else:
-            h = h
-            bins = None
-
-        if check_only:
-            return True if bins is not None else False
-        else:
-            return np.asarray(h).astype(float), np.asarray(bins).astype(float)
-
-    # Try to understand input
-    if type(H) is list:
-        # A list of objects
-        type_check = [hist_object_handler(h, check_only=True) for h in H]
-
-        if all(type_check):
-            # All objects are understandable
-            NH = len(type_check)
-            _H, _bins = [], []
-            for h in H:
-                h, bins = hist_object_handler(h)
-                _H.append(h)
-                _bins.append(bins)
-            if len(set(map(len, _H))) != 1:
-                raise ValueError(
-                    "Plotting multiple histograms with different binning is not supported"
-                )
-            h = np.asarray(_H).astype(float)
-            bins = _bins[0]
-        elif any(type_check):
-            # Some objects are not understandable
-            raise ValueError("Some hist objects were not recognized")
-        else:
-            # List of lists or arrays
-            h = np.asarray(H).astype(float)
-            bins = np.asarray(bins).astype(float) if bins is not None else None
-            NH = len(h) if h.ndim > 1 else 1
-    else:
-        # Single object, numpy array, 2D array
-        if hist_object_handler(H, check_only=True):
-            NH = 1
-            h, bins = hist_object_handler(H)
-        else:
-            # 2D array
-            h = np.asarray(H).astype(float)
-            bins = np.asarray(bins).astype(float) if bins is not None else None
-            NH = len(h) if h.ndim > 1 else 1
-
-    if bins is None and NH == 1:
-        bins = np.arange(len(h) + 1)
-    elif bins is None and NH > 1:
-        bins = [np.arange(len(_h) + 1) for _h in h][0]
+    # TODO: use hists everywhere
+    h = np.stack([h.values().astype(float) for h in hists])
+    if yerr is None and all(h.variances() is not None for h in hists):
+        yerr = np.stack([np.sqrt(h.variances()) for h in hists])
+    elif yerr is False:
+        yerr = None
 
     # Convert 1/0 etc to real bools
     stack = bool(stack)
     density = bool(density)
     edges = bool(edges)
     binticks = bool(binticks)
-    assert bins.ndim == 1, "bins need to be 1 dimensional"
-    # assert bins.shape[0] == h.shape[-1] + 1, (
-    #     "len along main axis of h has " "to be smaller by 1 than len " "of bins"
-    # )
-    assert w2 is None or yerr is None, "Can only supply errors or w2"
 
+    assert final_bins.ndim == 1, "bins need to be 1 dimensional"
+    # TODO: can check validity of bin length elsewhere
+
+    if w2 is not None and yerr is not None:
+        raise ValueError("Can only supply errors or w2")
+
+    _labels: List[Optional[str]]
     if label is None:
-        _labels = [None] * NH
+        _labels = [None] * len(hists)
     elif isinstance(label, str):
-        _labels = [label] * NH
+        _labels = [label] * len(hists)
     elif not np.iterable(label):
-        _labels = [str(label)] * NH
+        _labels = [str(label)] * len(hists)
     else:
         _labels = [str(lab) for lab in label]
 
     def iterable_not_string(arg):
         return isinstance(arg, collections.abc.Iterable) and not isinstance(arg, str)
 
-    _chunked_kwargs = []
-    for _i in range(NH):
+    _chunked_kwargs: List[Dict[str, Any]] = []
+    for _i in range(len(hists)):
         _chunked_kwargs.append({})
     for kwarg in kwargs:
         # Check if iterable
@@ -237,11 +202,12 @@ def histplot(
             for i in range(len(_chunked_kwargs)):
                 _chunked_kwargs[i][kwarg] = kwargs[kwarg]
 
-    _bin_widths = np.diff(bins)
-    _bin_centers = bins[1:] - _bin_widths / float(2)
+    _bin_widths = np.diff(final_bins)
+    _bin_centers = final_bins[1:] - _bin_widths / float(2)
 
     ############################
     # yerr calculation
+    _yerr: "np.ndarray | None"
     if yerr is not None:
         # yerr is array
         if hasattr(yerr, "__len__"):
@@ -252,17 +218,20 @@ def histplot(
         # yerr is automatic
         else:
             if yerr is True:
-                assert stack is False, (
-                    "Automatic errorbars not defined for " "stacked plot"
-                )
+                if stack:
+                    raise RuntimeError(
+                        "Automatic errorbars not defined for " "stacked plot"
+                    )
                 _yerr = np.sqrt(h)
+            else:
+                raise RuntimeError("Empty code path")
 
     elif w2 is not None:
         if w2method is None:
             int_w2 = np.around(w2).astype(int)
             if np.all(np.isclose(w2, int_w2, 0.000001)):
                 # If w2 are integers (true data hist), calculate Garwood interval
-                _yerr = np.abs(poisson_interval(h, w2) - h)
+                _yerr = np.abs(poisson_interval(h[0], w2) - h[0])
             else:
                 # w2 to errors directly if specified previously
                 _yerr = np.sqrt(w2)
@@ -273,87 +242,73 @@ def histplot(
         _yerr = None
 
     if _yerr is not None:
+        assert isinstance(_yerr, np.ndarray)
         if _yerr.ndim == 3:
             # Already correct format
             pass
-        elif _yerr.ndim == 2 and NH == 1:
+        elif _yerr.ndim == 2 and len(hists) == 1:
             # Broadcast ndim 2 to ndim 3
             if _yerr.shape[-2] == 2:  # [[1,1], [1,1]]
-                _yerr = _yerr.reshape(NH, 2, _yerr.shape[-1])
+                _yerr = _yerr.reshape(len(hists), 2, _yerr.shape[-1])
             elif _yerr.shape[-2] == 1:  # [[1,1]]
-                _yerr = np.tile(_yerr, 2).reshape(NH, 2, _yerr.shape[-1])
+                _yerr = np.tile(_yerr, 2).reshape(len(hists), 2, _yerr.shape[-1])
             else:
                 raise ValueError("yerr format is not understood")
         elif _yerr.ndim == 2:
             # Broadcast yerr (nh, N) to (nh, 2, N)
-            _yerr = np.tile(_yerr, 2).reshape(NH, 2, _yerr.shape[-1])
+            _yerr = np.tile(_yerr, 2).reshape(len(hists), 2, _yerr.shape[-1])
         elif _yerr.ndim == 1:
             # Broadcast yerr (1, N) to (nh, 2, N)
-            _yerr = np.tile(_yerr, 2 * NH).reshape(NH, 2, _yerr.shape[-1])
+            _yerr = np.tile(_yerr, 2 * len(hists)).reshape(
+                len(hists), 2, _yerr.shape[-1]
+            )
         else:
             raise ValueError("yerr format is not understood")
+
+        assert _yerr is not None
         # Split
         _yerr_lo = _yerr[:, 0]
         _yerr_hi = _yerr[:, 1]
 
     ############################
     # Stacking, norming, density
-    def get_stack(_h):
-        if NH > 1:
-            return np.cumsum(_h, axis=0)
-        else:
-            return _h
-
-    def get_density(h, density=True, binwnorm=None, bins=bins):
-        assert (not density) ^ (
-            binwnorm is None
-        ), "Can only calculate density or binwnorm"
-        per_hist_norm = np.sum(h, axis=(1 if h.ndim > 1 else 0))
-        if binwnorm is not None:
-            overallnorm = binwnorm * per_hist_norm
-        else:
-            overallnorm = np.ones(NH)
-        binnorms = np.outer(overallnorm, np.ones_like(bins[:-1]))
-        binnorms /= np.outer(np.diff(bins), per_hist_norm).T
-        if binnorms.ndim == 2 and len(binnorms) == 1:  # Unwrap if [[1,2,3]]
-            binnorms = binnorms[0]
-        return binnorms
 
     if density or binwnorm is not None:
-        density_arr = get_density(h, density, binwnorm)
+        density_arr = get_density(h, density, binwnorm, bins=final_bins, hists=hists)
         if stack:
-            h = get_stack(h)
-            h *= get_density(h, density, binwnorm)[-1]
+            h = get_stack(h, hists)
+            h *= get_density(h, density, binwnorm, bins=final_bins, hists=hists)[-1]
         else:
             h *= density_arr
         if _yerr is not None:
-            for i in range(NH):
+            for i in range(len(hists)):
                 _yerr_lo[i] = _yerr_lo[i] * density_arr[i]
                 _yerr_hi[i] = _yerr_hi[i] * density_arr[i]
 
     if stack and not density and binwnorm is None:
-        h = get_stack(h)
+        h = get_stack(h, hists)
 
     # Stack
-    if stack and NH > 1:
+    if stack and len(hists) > 1:
         h = h[::-1]
         _labels = _labels[::-1]
         _chunked_kwargs = _chunked_kwargs[::-1]
 
     #################################
     # Split and reshape for iteration
-    h = h.reshape(NH, -1)
+    h = h.reshape(len(hists), -1)
 
     ##########
     # Plotting
-    return_artists = []
+    return_artists: List[Union[FillArtists, StepArtists, ErrorBarArtists]] = []
     if histtype == "step":
-        for i in range(NH):
+        for i in range(len(hists)):
             if edges:
-                _bins = [bins[0], *bins, bins[-1]]
+                _bins = [final_bins[0], *final_bins, final_bins[-1]]
                 _h = [0, *h[i], h[i][-1], 0]
             else:
-                _bins, _h = bins, [*h[i], h[i][-1]]
+                _bins = list(final_bins)
+                _h = [*h[i], h[i][-1]]
             _kwargs = _chunked_kwargs[i]
             _label = _labels[i]
             _step_label = _label if yerr is None else None
@@ -383,10 +338,12 @@ def histplot(
         _artist = _s
 
     elif histtype == "fill":
-        for i in range(NH):
+        for i in range(len(hists)):
             _h = np.r_[h[i], h[i][-1]]
             _kwargs = _chunked_kwargs[i]
-            _f = ax.fill_between(bins, _h, step="post", label=_labels[i], **_kwargs)
+            _f = ax.fill_between(
+                final_bins, _h, step="post", label=_labels[i], **_kwargs
+            )
         return_artists.append(FillArtists(_f))
         _artist = _f
 
@@ -397,6 +354,7 @@ def histplot(
             "markersize": 10.0,
             "elinewidth": 1,
         }
+        _yerri: Optional[List[float]]
         for k, v in err_defaults.items():
             if k not in kwargs.keys():
                 kwargs[k] = v
@@ -404,7 +362,7 @@ def histplot(
                 if k not in kwarg_row.keys():
                     kwarg_row[k] = v
 
-        for i in range(NH):
+        for i in range(len(hists)):
             if yerr is not None or w2 is not None:
                 _yerri = [_yerr_lo[i], _yerr_hi[i]]
             else:
@@ -413,13 +371,15 @@ def histplot(
                 _bin_centers, h[i], yerr=_yerri, label=_labels[i], **_chunked_kwargs[i]
             )
             return_artists.append(ErrorBarArtists(_e))
+
         _artist = ax.plot(_bin_centers, h[0], lw=0, ms=0, alpha=0)[0]
+
     # Add sticky edges for autoscale
     _artist.sticky_edges.y.append(0)
 
     if binticks:
-        _slice = int(round(float(len(bins)) / len(ax.get_xticks()))) + 1
-        ax.set_xticks(bins[::_slice])
+        _slice = int(round(float(len(final_bins)) / len(ax.get_xticks()))) + 1
+        ax.set_xticks(final_bins[::_slice])
 
     if x_axes_label:
         ax.set_xlabel(x_axes_label)
@@ -436,6 +396,7 @@ def hist2dplot(
     cbarsize="7%",
     cbarpad=0.2,
     cbarpos="right",
+    cbarextend=False,
     cmin=None,
     cmax=None,
     ax=None,
@@ -470,6 +431,9 @@ def hist2dplot(
         Colorbar distance from main axis.
     cbarpos : {'right', 'left', 'bottom', 'top'}, optional,  default "right"
         Colorbar position w.r.t main axis.
+    cbarextend : bool, optional, default False
+        Extends figure size to keep original axes size same as without cbar.
+        Only safe for 1 axes per fig.
     cmin : float, optional
         Colorbar minimum.
     cmax : float, optional
@@ -485,49 +449,28 @@ def hist2dplot(
 
     """
 
+    # ax check
+    if ax is None:
+        ax = plt.gca()
+    else:
+        if not isinstance(ax, plt.Axes):
+            raise ValueError("ax must be a matplotlib Axes object")
+
     x_axes_label = ""
     y_axes_label = ""
 
-    # Input check
-    if hasattr(H, "axes") and hasattr(H, "to_numpy"):
-        # Hist compat
-        if len(H.axes) != 2:
-            raise ValueError("Not a 2D histogram")
+    hist = hist_object_handler(H, xbins, ybins)
 
-        x_axes_label = get_histogram_axes_title(H.axes[0])
-        y_axes_label = get_histogram_axes_title(H.axes[1])
+    # TODO: use Histogram everywhere
+    H = hist.values()
+    xbins = get_plottable_protocol_bins(hist.axes[0])
+    ybins = get_plottable_protocol_bins(hist.axes[1])
 
-        H, xbins, ybins = H.to_numpy()
-
-    elif hasattr(H, "to_numpy"):
-        # Generic (Uproot 4) support
-        H, xbins, ybins = H.to_numpy(flow=False)
-
-    elif hasattr(H, "numpy"):
-        # uproot/TH2
-        _tup = H.numpy()
-        if len(_tup) != 2:
-            raise ValueError("numpy() method not understood")
-        else:
-            H, xy = _tup
-            if len(xy) == 1 and isinstance(xy, list):
-                if len(xy[0]) == 2 and isinstance(xy[0], tuple):
-                    xbins, ybins = xy[0]
-                else:
-                    raise ValueError("Input not understood")
-            else:
-                raise ValueError("Input not understood")
+    x_axes_label = get_histogram_axes_title(hist.axes[0])
+    y_axes_label = get_histogram_axes_title(hist.axes[1])
 
     H = H.T
 
-    if ax is None:
-        ax = plt.gca()
-
-    if xbins is None:
-        xbins = np.arange(H.shape[1] + 1)
-    if ybins is None:
-        ybins = np.arange(H.shape[0] + 1)
-
     if cmin is not None:
         H[H < cmin] = None
     if cmax is not None:
@@ -552,7 +495,9 @@ def hist2dplot(
         ax.set_yticks(ybins)
 
     if cbar:
-        cax = append_axes(ax, size=cbarsize, pad=cbarpad, position=cbarpos)
+        cax = append_axes(
+            ax, size=cbarsize, pad=cbarpad, position=cbarpos, extend=cbarextend
+        )
         cb_obj = plt.colorbar(pc, cax=cax)
     else:
         cb_obj = None
@@ -665,8 +610,13 @@ def yscale_legend(ax=None):
     if ax is None:
         ax = plt.gca()
 
+    if ax.get_yscale() == "log":
+        scale_factor = 10 ** (1.05)
+    else:
+        scale_factor = 1.05
+
     while overlap(ax, _draw_leg_bbox(ax)) > 0:
-        ax.set_ylim(ax.get_ylim()[0], ax.get_ylim()[-1] * 1.05)
+        ax.set_ylim(ax.get_ylim()[0], ax.get_ylim()[-1] * scale_factor)
         ax.figure.canvas.draw()
     return ax
 
@@ -803,7 +753,7 @@ def make_square_add_cbar(ax, size=0.4, pad=0.1):
     return cax
 
 
-def append_axes(ax, size=0.1, pad=0.1, position="right"):
+def append_axes(ax, size=0.1, pad=0.1, position="right", extend=False):
     """
     Append a side ax to the current figure and return it.
     Figure is automatically extended along the direction of the added axes to
@@ -833,37 +783,43 @@ def append_axes(ax, size=0.1, pad=0.1, position="right"):
         xsizes = xsizes[::-1]
     yhax = divider.append_axes(position, size=margin_size, pad=pad_size)
 
-    def extend_ratio(ax):
-        ax.figure.canvas.draw()
-        orig_size = ax.get_position().size
-        new_size = 0
-        for itax in ax.figure.axes:
-            new_size += itax.get_position().size
+    if extend:
 
-        return new_size / orig_size
+        def extend_ratio(ax, yhax):
+            ax.figure.canvas.draw()
+            orig_size = ax.get_position().size
+            new_size = 0
+            for itax in [ax, yhax]:
+                new_size += itax.get_position().size
 
-    if position in ["right"]:
-        divider.set_horizontal([axes_size.Fixed(width)] + xsizes)
-        fig.set_size_inches(
-            fig.get_size_inches()[0] * extend_ratio(ax)[0], fig.get_size_inches()[1]
-        )
-    elif position in ["left"]:
-        divider.set_horizontal(xsizes[::-1] + [axes_size.Fixed(width)])
-        fig.set_size_inches(
-            fig.get_size_inches()[0] * extend_ratio(ax)[0], fig.get_size_inches()[1]
-        )
-    elif position in ["top"]:
-        divider.set_vertical([axes_size.Fixed(height)] + xsizes[::-1])
-        fig.set_size_inches(
-            fig.get_size_inches()[0], fig.get_size_inches()[1] * extend_ratio(ax)[1]
-        )
-        ax.get_shared_x_axes().join(ax, yhax)
-    elif position in ["bottom"]:
-        divider.set_vertical(xsizes + [axes_size.Fixed(height)])
-        fig.set_size_inches(
-            fig.get_size_inches()[0], fig.get_size_inches()[1] * extend_ratio(ax)[1]
-        )
-        ax.get_shared_x_axes().join(ax, yhax)
+            return new_size / orig_size
+
+        if position in ["right"]:
+            divider.set_horizontal([axes_size.Fixed(width)] + xsizes)
+            fig.set_size_inches(
+                fig.get_size_inches()[0] * extend_ratio(ax, yhax)[0],
+                fig.get_size_inches()[1],
+            )
+        elif position in ["left"]:
+            divider.set_horizontal(xsizes[::-1] + [axes_size.Fixed(width)])
+            fig.set_size_inches(
+                fig.get_size_inches()[0] * extend_ratio(ax, yhax)[0],
+                fig.get_size_inches()[1],
+            )
+        elif position in ["top"]:
+            divider.set_vertical([axes_size.Fixed(height)] + xsizes[::-1])
+            fig.set_size_inches(
+                fig.get_size_inches()[0],
+                fig.get_size_inches()[1] * extend_ratio(ax, yhax)[1],
+            )
+            ax.get_shared_x_axes().join(ax, yhax)
+        elif position in ["bottom"]:
+            divider.set_vertical(xsizes + [axes_size.Fixed(height)])
+            fig.set_size_inches(
+                fig.get_size_inches()[0],
+                fig.get_size_inches()[1] * extend_ratio(ax, yhax)[1],
+            )
+            ax.get_shared_x_axes().join(ax, yhax)
 
     return yhax
 
@@ -893,7 +849,6 @@ def sort_legend(ax, order=None):
     ax : axes with legend labels in it
     order : Ordered dict with renames or array with order
     """
-    from collections import OrderedDict
 
     handles, labels = ax.get_legend_handles_labels()
     by_label = OrderedDict(zip(labels, handles))
@@ -901,7 +856,7 @@ def sort_legend(ax, order=None):
     if isinstance(order, OrderedDict):
         ordered_label_list = list(order.keys())
     elif isinstance(order, (list, tuple, np.ndarray)):
-        ordered_label_list = order
+        ordered_label_list = list(order)
     elif order is None:
         ordered_label_list = labels
     else:
